@@ -13,13 +13,15 @@
 # under the License.
 
 import importlib
+import math
 import warnings
 
 from keystoneauth1 import adapter
-from keystoneauth1 import plugin
+import keystoneauth1.exceptions.catalog
 from keystoneauth1 import session
 import requestsexceptions
 
+import os_client_config
 from os_client_config import _log
 from os_client_config import constructors
 from os_client_config import exceptions
@@ -70,7 +72,8 @@ def _make_key(key, service_type):
 class CloudConfig(object):
     def __init__(self, name, region, config,
                  force_ipv4=False, auth_plugin=None,
-                 openstack_config=None):
+                 openstack_config=None, session_constructor=None,
+                 app_name=None, app_version=None):
         self.name = name
         self.region = region
         self.config = config
@@ -79,6 +82,9 @@ class CloudConfig(object):
         self._auth = auth_plugin
         self._openstack_config = openstack_config
         self._keystone_session = None
+        self._session_constructor = session_constructor or session.Session
+        self._app_name = app_name
+        self._app_version = app_version
 
     def __getattr__(self, key):
         """Return arbitrary attributes."""
@@ -101,6 +107,10 @@ class CloudConfig(object):
     def __ne__(self, other):
         return not self == other
 
+    def set_session_constructor(self, session_constructor):
+        """Sets the Session constructor."""
+        self._session_constructor = session_constructor
+
     def get_requests_verify_args(self):
         """Return the verify and cert values for the requests library."""
         if self.config['verify'] and self.config['cacert']:
@@ -154,9 +164,16 @@ class CloudConfig(object):
         # atrocity from the as-yet-unsullied eyes of our users.
         # Of course, if the user requests a volumev2, that structure should
         # still work.
-        if (service_type == 'volume' and
-                self.get_api_version(service_type).startswith('2')):
-            service_type = 'volumev2'
+        # What's even more amazing is that they did it AGAIN with cinder v3
+        # And then I learned that mistral copied it.
+        if service_type == 'volume':
+            if self.get_api_version(service_type).startswith('2'):
+                service_type = 'volumev2'
+            elif self.get_api_version(service_type).startswith('3'):
+                service_type = 'volumev3'
+        elif service_type == 'workflow':
+            if self.get_api_version(service_type).startswith('2'):
+                service_type = 'workflowv2'
         return self.config.get(key, service_type)
 
     def get_service_name(self, service_type):
@@ -196,13 +213,28 @@ class CloudConfig(object):
                     " since verify=False".format(
                         cloud=self.name, region=self.region))
             requestsexceptions.squelch_warnings(insecure_requests=not verify)
-            self._keystone_session = session.Session(
+            self._keystone_session = self._session_constructor(
                 auth=self._auth,
                 verify=verify,
                 cert=cert,
                 timeout=self.config['api_timeout'])
+            if hasattr(self._keystone_session, 'additional_user_agent'):
+                self._keystone_session.additional_user_agent.append(
+                    ('os-client-config', os_client_config.__version__))
+            # Using old keystoneauth with new os-client-config fails if
+            # we pass in app_name and app_version. Those are not essential,
+            # nor a reason to bump our minimum, so just test for the session
+            # having the attribute post creation and set them then.
+            if hasattr(self._keystone_session, 'app_name'):
+                self._keystone_session.app_name = self._app_name
+            if hasattr(self._keystone_session, 'app_version'):
+                self._keystone_session.app_version = self._app_version
         return self._keystone_session
 
+    def get_service_catalog(self):
+        """Helper method to grab the service catalog."""
+        return self._auth.get_access(self.get_session()).service_catalog
+
     def get_session_client(self, service_key):
         """Return a prepped requests adapter for a given service.
 
@@ -225,7 +257,19 @@ class CloudConfig(object):
             interface=self.get_interface(service_key),
             region_name=self.region)
 
-    def get_session_endpoint(self, service_key):
+    def _get_highest_endpoint(self, service_types, kwargs):
+        session = self.get_session()
+        for service_type in service_types:
+            kwargs['service_type'] = service_type
+            try:
+                # Return the highest version we find that matches
+                # the request
+                return session.get_endpoint(**kwargs)
+            except keystoneauth1.exceptions.catalog.EndpointNotFound:
+                pass
+
+    def get_session_endpoint(
+            self, service_key, min_version=None, max_version=None):
         """Return the endpoint from config or the catalog.
 
         If a configuration lists an explicit endpoint for a service,
@@ -235,27 +279,49 @@ class CloudConfig(object):
         :param service_key: Generic key for service, such as 'compute' or
                             'network'
 
-        :returns: Endpoint for the service, or None if not found
         """
 
         override_endpoint = self.get_endpoint(service_key)
         if override_endpoint:
             return override_endpoint
-        # keystone is a special case in keystone, because what?
-        session = self.get_session()
-        if service_key == 'identity':
-            endpoint = session.get_endpoint(interface=plugin.AUTH_INTERFACE)
+        endpoint = None
+        kwargs = {
+            'service_name': self.get_service_name(service_key),
+            'region_name': self.region
+        }
+        kwargs['interface'] = self.get_interface(service_key)
+        if service_key == 'volume' and not self.get_api_version('volume'):
+            # If we don't have a configured cinder version, we can't know
+            # to request a different service_type
+            min_version = float(min_version or 1)
+            max_version = float(max_version or 3)
+            min_major = math.trunc(float(min_version))
+            max_major = math.trunc(float(max_version))
+            versions = range(int(max_major) + 1, int(min_major), -1)
+            service_types = []
+            for version in versions:
+                if version == 1:
+                    service_types.append('volume')
+                else:
+                    service_types.append('volumev{v}'.format(v=version))
         else:
-            endpoint = session.get_endpoint(
-                service_type=self.get_service_type(service_key),
-                service_name=self.get_service_name(service_key),
-                interface=self.get_interface(service_key),
-                region_name=self.region)
+            service_types = [self.get_service_type(service_key)]
+        endpoint = self._get_highest_endpoint(service_types, kwargs)
+        if not endpoint:
+            self.log.warning(
+                "Keystone catalog entry not found ("
+                "service_type=%s,service_name=%s"
+                "interface=%s,region_name=%s)",
+                service_key,
+                kwargs['service_name'],
+                kwargs['interface'],
+                kwargs['region_name'])
         return endpoint
 
     def get_legacy_client(
             self, service_key, client_class=None, interface_key=None,
-            pass_version_arg=True, version=None, **kwargs):
+            pass_version_arg=True, version=None, min_version=None,
+            max_version=None, **kwargs):
         """Return a legacy OpenStack client object for the given config.
 
         Most of the OpenStack python-*client libraries have the same
@@ -289,6 +355,8 @@ class CloudConfig(object):
                                  that case.
         :param version: (optional) Version string to override the configured
                                    version string.
+        :param min_version: (options) Minimum version acceptable.
+        :param max_version: (options) Maximum version acceptable.
         :param kwargs: (optional) keyword args are passed through to the
                        Client constructor, so this is in case anything
                        additional needs to be passed in.
@@ -296,26 +364,26 @@ class CloudConfig(object):
         if not client_class:
             client_class = _get_client(service_key)
 
-        # Because of course swift is different
-        if service_key == 'object-store':
-            return self._get_swift_client(client_class=client_class, **kwargs)
         interface = self.get_interface(service_key)
         # trigger exception on lack of service
-        endpoint = self.get_session_endpoint(service_key)
+        endpoint = self.get_session_endpoint(
+            service_key, min_version=min_version, max_version=max_version)
         endpoint_override = self.get_endpoint(service_key)
 
-        if not interface_key:
-            if service_key in ('image', 'key-manager'):
-                interface_key = 'interface'
-            else:
-                interface_key = 'endpoint_type'
-
-        constructor_kwargs = dict(
-            session=self.get_session(),
-            service_name=self.get_service_name(service_key),
-            service_type=self.get_service_type(service_key),
-            endpoint_override=endpoint_override,
-            region_name=self.region)
+        if service_key == 'object-store':
+            constructor_kwargs = dict(
+                session=self.get_session(),
+                os_options=dict(
+                    service_type=self.get_service_type(service_key),
+                    object_storage_url=endpoint_override,
+                    region_name=self.region))
+        else:
+            constructor_kwargs = dict(
+                session=self.get_session(),
+                service_name=self.get_service_name(service_key),
+                service_type=self.get_service_type(service_key),
+                endpoint_override=endpoint_override,
+                region_name=self.region)
 
         if service_key == 'image':
             # os-client-config does not depend on glanceclient, but if
@@ -334,10 +402,12 @@ class CloudConfig(object):
             if not endpoint_override:
                 constructor_kwargs['endpoint_override'] = endpoint
         constructor_kwargs.update(kwargs)
-        constructor_kwargs[interface_key] = interface
-        if pass_version_arg:
+        if pass_version_arg and service_key != 'object-store':
             if not version:
                 version = self.get_api_version(service_key)
+            if not version and service_key == 'volume':
+                from cinderclient import client as cinder_client
+                version = cinder_client.get_volume_api_from_url(endpoint)
             # Temporary workaround while we wait for python-openstackclient
             # to be able to handle 2.0 which is what neutronclient expects
             if service_key == 'network' and version == '2':
@@ -349,61 +419,47 @@ class CloudConfig(object):
                     constructor_kwargs['endpoint'] = endpoint
             if service_key == 'network':
                 constructor_kwargs['api_version'] = version
+            elif service_key == 'baremetal':
+                if version != '1':
+                    # Set Ironic Microversion
+                    constructor_kwargs['os_ironic_api_version'] = version
+                # Version arg is the major version, not the full microstring
+                constructor_kwargs['version'] = version[0]
             else:
                 constructor_kwargs['version'] = version
+            if min_version and min_version > float(version):
+                raise exceptions.OpenStackConfigVersionException(
+                    "Minimum version {min_version} requested but {version}"
+                    " found".format(min_version=min_version, version=version),
+                    version=version)
+            if max_version and max_version < float(version):
+                raise exceptions.OpenStackConfigVersionException(
+                    "Maximum version {max_version} requested but {version}"
+                    " found".format(max_version=max_version, version=version),
+                    version=version)
+        if service_key == 'database':
+            # TODO(mordred) Remove when https://review.openstack.org/314032
+            # has landed and released. We're passing in a Session, but the
+            # trove Client object has username and password as required
+            # args
+            constructor_kwargs['username'] = None
+            constructor_kwargs['password'] = None
+
+        if not interface_key:
+            if service_key in ('image', 'key-manager'):
+                interface_key = 'interface'
+            elif (service_key == 'identity'
+                  and version and version.startswith('3')):
+                interface_key = 'interface'
+            else:
+                interface_key = 'endpoint_type'
+        if service_key == 'object-store':
+            constructor_kwargs['os_options'][interface_key] = interface
+        else:
+            constructor_kwargs[interface_key] = interface
 
         return client_class(**constructor_kwargs)
 
-    def _get_swift_client(self, client_class, **kwargs):
-        auth_args = self.get_auth_args()
-        auth_version = self.get_api_version('identity')
-        session = self.get_session()
-        token = session.get_token()
-        endpoint = self.get_session_endpoint(service_key='object-store')
-        if not endpoint:
-            return None
-        # If we have a username/password, we want to pass them to
-        # swift - because otherwise it will not re-up tokens appropriately
-        # However, if we only have non-password auth, then get a token
-        # and pass it in
-        swift_kwargs = dict(
-            auth_version=auth_version,
-            preauthurl=endpoint,
-            preauthtoken=token,
-            os_options=dict(
-                region_name=self.get_region_name(),
-                auth_token=token,
-                object_storage_url=endpoint,
-                service_type=self.get_service_type('object-store'),
-                endpoint_type=self.get_interface('object-store'),
-
-            ))
-        if self.config['api_timeout'] is not None:
-            swift_kwargs['timeout'] = float(self.config['api_timeout'])
-
-        # create with password
-        swift_kwargs['user'] = auth_args.get('username')
-        swift_kwargs['key'] = auth_args.get('password')
-        swift_kwargs['authurl'] = auth_args.get('auth_url')
-        os_options = {}
-        if auth_version == '2.0':
-            os_options['tenant_name'] = auth_args.get('project_name')
-            os_options['tenant_id'] = auth_args.get('project_id')
-        else:
-            os_options['project_name'] = auth_args.get('project_name')
-            os_options['project_id'] = auth_args.get('project_id')
-
-        for key in (
-                'user_id',
-                'project_domain_id',
-                'project_domain_name',
-                'user_domain_id',
-                'user_domain_name'):
-            os_options[key] = auth_args.get(key)
-        swift_kwargs['os_options'].update(os_options)
-
-        return client_class(**swift_kwargs)
-
     def get_cache_expiration_time(self):
         if self._openstack_config:
             return self._openstack_config.get_cache_expiration_time()
@@ -438,3 +494,65 @@ class CloudConfig(object):
             if resource not in expiration:
                 return default
             return float(expiration[resource])
+
+    def requires_floating_ip(self):
+        """Return whether or not this cloud requires floating ips.
+
+
+        :returns: True of False if know, None if discovery is needed.
+                  If requires_floating_ip is not configured but the cloud is
+                  known to not provide floating ips, will return False.
+        """
+        if self.config['floating_ip_source'] == "None":
+            return False
+        return self.config['requires_floating_ip']
+
+    def get_external_networks(self):
+        """Get list of network names for external networks."""
+        return [
+            net['name'] for net in self.config['networks']
+            if net['routes_externally']]
+
+    def get_external_ipv4_networks(self):
+        """Get list of network names for external IPv4 networks."""
+        return [
+            net['name'] for net in self.config['networks']
+            if net['routes_ipv4_externally']]
+
+    def get_external_ipv6_networks(self):
+        """Get list of network names for external IPv6 networks."""
+        return [
+            net['name'] for net in self.config['networks']
+            if net['routes_ipv6_externally']]
+
+    def get_internal_networks(self):
+        """Get list of network names for internal networks."""
+        return [
+            net['name'] for net in self.config['networks']
+            if not net['routes_externally']]
+
+    def get_internal_ipv4_networks(self):
+        """Get list of network names for internal IPv4 networks."""
+        return [
+            net['name'] for net in self.config['networks']
+            if not net['routes_ipv4_externally']]
+
+    def get_internal_ipv6_networks(self):
+        """Get list of network names for internal IPv6 networks."""
+        return [
+            net['name'] for net in self.config['networks']
+            if not net['routes_ipv6_externally']]
+
+    def get_default_network(self):
+        """Get network used for default interactions."""
+        for net in self.config['networks']:
+            if net['default_interface']:
+                return net['name']
+        return None
+
+    def get_nat_destination(self):
+        """Get network used for NAT destination."""
+        for net in self.config['networks']:
+            if net['nat_destination']:
+                return net['name']
+        return None
